#ifndef FS_TYPE_H
#define FS_TYPE_H

#include "const.h"
#include "sys/type.h"

// inode memorizzato su disco

// Declaration of the V1 inode as it is on the disk (not in core). 
struct d1_inode {		/* V1.x disk inode */
	mode_t d1_mode;		/* file type, protection, etc. */
	uid_t d1_uid;			/* user id of the file's owner */
	off_t d1_size;		/* current file size in bytes */
	time_t d1_mtime;		/* when was file data last changed */
	uint8_t d1_gid;			/* group number */
	uint8_t d1_nlinks;		/* how many links to this file */
	uint16_t d1_zone[V1_NR_TZONES];	/* block nums for direct, ind, and dbl ind */
} __attribute__((packed));

// Declaration of the V2 inode as it is on the disk (not in core). 
struct d2_inode {		/* V2.x disk inode */
	mode_t d2_mode;		/* file type, protection, etc. */
	uint16_t d2_nlinks;		/* how many links to this file. HACK! */
	uid_t d2_uid;			/* user id of the file's owner. */
	uint16_t d2_gid;			/* group number HACK! */
	off_t d2_size;		/* current file size in bytes */
	time_t d2_atime;		/* when was file data last accessed */
	time_t d2_mtime;		/* when was file data last changed */
	time_t d2_ctime;		/* when was inode data last changed */
	zone_t d2_zone[V2_NR_TZONES];	/* block nums for direct, ind, and dbl ind */
} __attribute__((packed));

/* Inode table.  This table holds inodes that are currently in use.  In some
 * cases they have been opened by an open() or creat() system call, in other
 * cases the file system itself needs the inode for one reason or another,
 * such as to search a directory for a path name.
 * The first part of the struct holds fields that are present on the
 * disk; the second part holds fields not present on the disk.
 * The disk inode part is also declared in "type.h" as 'd1_inode' for V1
 * file systems and 'd2_inode' for V2 file systems.
 */

struct inode {
	mode_t i_mode;		/* file type, protection, etc. */
	nlink_t i_nlinks;		/* how many links to this file */
	uid_t i_uid;			/* user id of the file's owner */
	gid_t i_gid;			/* group number */
	off_t i_size;			/* current file size in bytes */
	time_t i_atime;		/* time of last access (V2 only) */
	time_t i_mtime;		/* when was file data last changed */
	time_t i_ctime;		/* when was inode itself changed (V2 only)*/
	zone_t i_zone[NR_TZONES]; /* zone numbers for direct, ind, and dbl ind */
	/* The following items are not present on the disk. */
	dev_t i_dev;			/* which device is the inode on */
	ino_t i_num;			/* inode number on its (minor) device */
	int i_count;			/* # times inode used; 0 means slot is free */
	int i_ndzones;		/* # direct zones (Vx_NR_DZONES) */
	int i_nindirs;		/* # indirect zones per indirect block */
	struct super_block *i_sp;	/* pointer to super block for inode's device */
	char i_dirt;			/* CLEAN or DIRTY */
	char i_pipe;			/* set to I_PIPE if pipe */
	char i_mount;			/* this bit is set if file mounted on */
	char i_seek;			/* set on LSEEK, cleared on READ/WRITE */
	char i_update;		/* the ATIME, CTIME, and MTIME bits are here */
};

/* Field values.  Note that CLEAN and DIRTY are defined in "const.h" */
#define NO_PIPE            0	/* i_pipe is NO_PIPE if inode is not a pipe */
#define I_PIPE             1	/* i_pipe is I_PIPE if inode is a pipe */
#define NO_MOUNT           0	/* i_mount is NO_MOUNT if file not mounted on*/
#define I_MOUNT            1	/* i_mount is I_MOUNT if file mounted on */
#define NO_SEEK            0	/* i_seek = NO_SEEK if last op was not SEEK */
#define ISEEK              1	/* i_seek = ISEEK if last op was SEEK */

#define NIL_INODE (struct inode *) 0    /* indicates absence of inode slot */

#endif
