#include "libce.h"
#include "sys/stat.h"
#include "sys/dir.h"
#include "super.h"
#include "file.h"
#include "fs.h"
#include "glo.h"
#include "inode.h"
#include "sys/fcntl.h"

PRIVATE char mode_map[] = {R_BIT, W_BIT, R_BIT|W_BIT, 0};
PRIVATE inode *new_node(char *path, mode_t bits, zone_t z0);
PRIVATE int common_open(const char *pathname, int oflags, mode_t omode);

/*===========================================================================*
 *				do_creat				     *
 *===========================================================================*/
PUBLIC int do_creat(const char *pathname, mode_t mode)
{
	// Perform the creat(name, mode) system call. 
	flog(LOG_DEBUG,"eseguo creat %s",pathname);
	int r = common_open(pathname, O_WRONLY | O_CREAT | O_TRUNC, mode);
	return(r);
}

/*===========================================================================*
 *				do_open					     *
 *===========================================================================*/
PUBLIC int do_open(const char *pathname, int flags, mode_t mode)
{
	// Perform the open(name, flags,...) system call. 
	int create_mode = 0;		// is really mode_t but this gives problems 
	int r;

    // If O_CREAT is set, open has three parameters, otherwise two. 
	if (mode & O_CREAT) 
		create_mode = mode;	
 
	r = common_open(pathname, flags, create_mode);
	return(r);
}

/*===========================================================================*
 *				common_open				     *
 *===========================================================================*/
PRIVATE int common_open(const char *pathname, int oflags, mode_t omode)
{
// Common code from do_creat and do_open. 
	//flog(LOG_DEBUG,"c_fs_open pathname:%s ",pathname);
	inode *rip;
	int r, exist = TRUE;
	mode_t bits;
	filp *fil_ptr;
	fproc *fp = get_fproc(getpid());
	int fd;

  	// Remap the bottom two bits of oflags. 
	bits = (mode_t) mode_map[oflags & O_ACCMODE];

  	// See if file descriptor and filp slots are available. 
	if ( (r = get_fd(0, bits, &fd, &fil_ptr)) != OK) return(r);

  	//flog(LOG_DEBUG,"do_open fd:%d",fd);
  	//If O_CREATE is set, try to make the file.  
  	if (oflags & O_CREAT) {
  	// Create a new inode by calling new_node(). 
		omode = I_REGULAR | (omode & ALL_MODES & fp->fp_umask);
    	rip = new_node((char*)pathname, omode, NO_ZONE);
    	r = err_code;
    	if (r == OK) exist = FALSE;      // we just created the file 
		else if (r != EEXIST) return(r); // other error 
		else exist = !(oflags & O_EXCL); /* file exists, if the O_EXCL 
					    					flag is set this is an error */
  	} else {
	 	// Scan path name. 
    	if ( (rip = eat_path((char*)pathname)) == NIL_INODE) return(err_code);
  	}

  	// Claim the file descriptor and filp slot and fill them in. 
	fp->fp_filp[fd] = fil_ptr;
	fil_ptr->filp_count = 1;
	fil_ptr->filp_ino = rip;
	fil_ptr->filp_flags = oflags;

  	// Only do the normal open code if we didn't just create the file. 
  	if (exist) {
		flog(LOG_DEBUG,"do_open - il file esiste");
		// Check protections. 
		if ((r = forbidden(rip, bits)) == OK) {
			flog(LOG_DEBUG,"do_open - accesso permesso");
			// Opening reg. files directories and special files differ. 
			switch (rip->i_mode & I_TYPE) {
				case I_REGULAR: 
				// Truncate regular file if O_TRUNC. 
					if (oflags & O_TRUNC) {
						if ((r = forbidden(rip, W_BIT)) !=OK) break;
							truncate(rip);
							wipe_inode(rip);
							/* Send the inode from the inode cache to the
							* block cache, so it gets written on the next
							* cache flush.
							*/
							rw_inode(rip, WRITING);
					}
					break;

				case I_DIRECTORY: 
					// Directories may be read but not written. 
					r = (bits & W_BIT ? EISDIR : OK);
					break;
			}
		}
  	}

  	// If error, release inode. 
	if (r != OK) {
		//if (r == SUSPEND) return(r);		// Oops, just suspended 
		fp->fp_filp[fd] = NIL_FILP;
		fil_ptr->filp_count= 0;
		put_inode(rip);
		return(r);
	}

	return fd;
}

/*===========================================================================*
 *				new_node				     *
 *===========================================================================*/
PRIVATE inode *new_node(char *path, mode_t bits, zone_t z0)
{
/* New_node() is called by common_open(), do_mknod(), and do_mkdir().  
 * In all cases it allocates a new inode, makes a directory entry for it on 
 * the path 'path', and initializes it.  It returns a pointer to the inode if 
 * it can do this; otherwise it returns NIL_INODE.  It always sets 'err_code'
 * to an appropriate value (OK or an error code).
 */

	inode *rlast_dir_ptr, *rip;
	int r;
	char string[NAME_MAX];

  	/* See if the path can be opened down to the last directory. */
  	if ((rlast_dir_ptr = last_dir(path, string)) == NIL_INODE) return(NIL_INODE);

	/* The final directory is accessible. Get final component of the path. */
  	rip = advance(rlast_dir_ptr, string);
  	if ( rip == NIL_INODE && err_code == ENOENT) {
		/* Last path component does not exist.  Make new directory entry. */
		if ( (rip = alloc_inode(rlast_dir_ptr->i_dev, bits)) == NIL_INODE) {
			/* Can't creat new inode: out of inodes. */
			put_inode(rlast_dir_ptr);
			return(NIL_INODE);
		}

		/* Force inode to the disk before making directory entry to make
		 * the system more robust in the face of a crash: an inode with
		 * no directory entry is much better than the opposite.
		 */

		flog(LOG_DEBUG,"new_node i_num:%d i_zone[0]:%d z0:%d",rip->i_num,rip->i_zone[0],z0);
		rip->i_nlinks++;
		rip->i_zone[0] = z0;		// major/minor device numbers 
		rw_inode(rip, WRITING);		// force inode to disk now 

		//flog(LOG_DEBUG,"inode scritto su disco i_num:%d creo entry directory",rip->i_num);
		//flog(LOG_DEBUG,"filename: %s",string);
		/* New inode acquired.  Try to make directory entry. */
		if ((r = search_dir(rlast_dir_ptr, string, &rip->i_num,ENTER)) != OK) {
			//flog(LOG_DEBUG,"directory non trovata");
			put_inode(rlast_dir_ptr);
			rip->i_nlinks--;	/* pity, have to free disk inode */
			rip->i_dirt = DIRTY;	/* dirty inodes are written out */
			put_inode(rip);	/* this call frees the inode */
			err_code = r;
			return(NIL_INODE);
		}
	} else {
		/* Either last component exists, or there is some problem. */
		if (rip != NIL_INODE)
			r = EEXIST;
		else
			r = err_code;
  	}

	/* Return the directory inode and exit. */
	put_inode(rlast_dir_ptr);
	//flog(LOG_DEBUG,"rlast_dir_ptr->i_num:%d rlast_dir_ptr->i_size:%d",rlast_dir_ptr->i_num,rlast_dir_ptr->i_size);
	err_code = r;
	return(rip);
}

/*===========================================================================*
 *				do_mkdir				     *
 *===========================================================================*/
PUBLIC int do_mkdir(const char *name, mode_t mode)
{
	// Perform the mkdir(name, mode) system call. 

	int r1, r2;			/* status codes */
	ino_t dot, dotdot;		/* inode numbers for . and .. */
	mode_t bits;			/* mode bits for the new inode */
	char string[NAME_MAX];	/* last component of the new dir's path name */
	inode *rip, *ldirp;
	fproc *fp = get_fproc(getpid());

	flog(LOG_DEBUG, "do_mkdir:%s",name);
  	// Check to see if it is possible to make another link in the parent dir. */
  	ldirp = last_dir(name, string);	// pointer to new dir's parent 
	if (ldirp == NIL_INODE) return(err_code);
	if (ldirp->i_nlinks >= (ldirp->i_sp->s_version == V1 ? CHAR_MAX : SHRT_MAX)) {
		put_inode(ldirp);	/* return parent */
		return(EMLINK);
  	}

  	// Next make the inode. If that fails, return error code. 
	bits = I_DIRECTORY | (mode & RWX_MODES & fp->fp_umask);
	rip = new_node(const_cast<char*>(name), bits, (zone_t) 0);
	if (rip == NIL_INODE || err_code == EEXIST) {
		put_inode(rip);		// can't make dir: it already exists 
		put_inode(ldirp);	// return parent too 
		return(err_code);
	}

  	// Get the inode numbers for . and .. to enter in the directory. 
	dotdot = ldirp->i_num;	// parent's inode number 
	dot = rip->i_num;		// inode number of the new dir itself 

  	// Now make dir entries for . and .. unless the disk is completely full. 
  	// Use dot1 and dot2, so the mode of the directory isn't important. 
	rip->i_mode = bits;	// set mode 
	r1 = search_dir(rip, dot1, &dot, ENTER);	/* enter . in the new dir */
  	r2 = search_dir(rip, dot2, &dotdot, ENTER);	/* enter .. in the new dir */

  	// If both . and .. were successfully entered, increment the link counts. 
	if (r1 == OK && r2 == OK) {
		// Normal case.  It was possible to enter . and .. in the new dir. 
		rip->i_nlinks++;	// this accounts for . 
		ldirp->i_nlinks++;	// this accounts for .. 
		ldirp->i_dirt = DIRTY;	// mark parent's inode as dirty 
  	} else {
		// It was not possible to enter . or .. probably disk was full. 
		(void) search_dir(ldirp, string, (ino_t *) 0, DELETE);
		rip->i_nlinks--;	// undo the increment done in new_node() 
  	}
  	rip->i_dirt = DIRTY;		// either way, i_nlinks has changed

	put_inode(ldirp);		// return the inode of the parent dir 
	put_inode(rip);		// return the inode of the newly made dir 
	return(err_code);		// new_node() always sets 'err_code'
}

/*===========================================================================*
 *				do_close				     *
 *===========================================================================*/
PUBLIC int do_close(int fd)
{
/* Perform the close(fd) system call. */

	filp *rfilp;
	inode *rip;
  	fproc *fp = get_fproc(getpid());

  	/* First locate the inode that belongs to the file descriptor. */
  	if ( (rfilp = get_filp(fd)) == NIL_FILP) return(err_code);

  	// mutua esclusione sul file
  	
  	rip = rfilp->filp_ino;	/* 'rip' points to the inode */

	/* If a write has been done, the inode is already marked as DIRTY. */
	if (--rfilp->filp_count == 0) {
		put_inode(rip);
	}

	fp->fp_cloexec &= ~(1L << fd);	// turn off close-on-exec bit 
	//flog(LOG_DEBUG,"!!!|do-close");
	fp->fp_filp[fd] = NIL_FILP;
	
	return(OK);
}