// sorgente adattato in C++ dai sorgenti ufficiali di minix

/* This file contains the procedures that look up path names in the directory
 * system and determine the inode number that goes with a given path name.
 *
 *  The entry points into this file are
 *   eat_path:	 the 'main' routine of the path-to-inode conversion mechanism
 *   last_dir:	 find the final directory on a given path
 *   advance:	 parse one component of a path name
 *   search_dir: search a directory for a string and return its inode number
 */
#include "libce.h"
#include "fs.h"
//#include <minix/callnr.h>
//#include "buf.h"
#include "file.h"
#include "fproc.h"
#include "inode.h"
#include "super.h"
#include "glo.h"
#include "blocco_fis.h"

char dot1[2] = ".";	/* used for search_dir to bypass the access */
char dot2[3] = "..";	/* permissions for . and ..		    */

extern int err_code;		
extern int rdwt_err;	

PRIVATE char *get_name(const char *old_name, char string[NAME_MAX]);

/*===========================================================================*
 *				eat_path				     *
 *===========================================================================*/
PUBLIC
inode *eat_path(const char *path)	/* the path name to be parsed */
{ 
/* Parse the path 'path' and put its inode in the inode table. If not possible,
 * return NIL_INODE as function value and an error code in 'err_code'.
 */

	inode *ldip, *rip;
	char string[NAME_MAX];	/* hold 1 path component name here */

	/* First open the path down to the final directory. */
	if ( (ldip = last_dir(path, string)) == NIL_INODE) {
		flog(LOG_DEBUG,"impossibile aprire cartella finale");
	return(NIL_INODE);	/* we couldn't open final directory */
	}

	/* The path consisting only of "/" is a special case, check for it. */
	if (string[0] == '\0') return(ldip);

	//flog(LOG_DEBUG,"!eat_path string: %s",string);
  /* Get final component of the path. */
  	rip = advance(ldip, string);
	put_inode(ldip);

	//flog(LOG_DEBUG,"!eat_path rip i_nlinks:%d i_num:%d",rip->i_nlinks,rip->i_num);
  //flog(LOG_DEBUG,"ldip i_nlinks:%d i_num:%d",ldip->i_nlinks,ldip->i_num);
	return(rip);
}

/*===========================================================================*
 *				last_dir				     *
 *===========================================================================*/

//char *path;			/* the path name to be parsed */
//char string[NAME_MAX];		/* the final component is returned here */
PUBLIC
inode *last_dir(const char *path, char string[NAME_MAX])
{
/* Given a path, 'path', located in the fs address space, parse it as
 * far as the last directory, fetch the inode for the last directory into
 * the inode table, and return a pointer to the inode.  In
 * addition, return the final component of the path in 'string'.
 * If the last directory can't be opened, return NIL_INODE and
 * the reason for failure in 'err_code'.
 */

  inode *rip;
  char *new_name;
  inode *new_ip;
  // in minix il filesystem è un server ed il descrittore fproc lo ottiene
  // dal messaggio. Il nostro kernel invece è monolitico
  fproc *fp = get_fproc(getpid()); 

  /* Is the path absolute or relative?  Initialize 'rip' accordingly. */
  rip = (*path == '/' ? fp->fp_rootdir : fp->fp_workdir);

  /* If dir has been removed or path is empty, return ENOENT. */
  if (rip->i_nlinks == 0 || *path == '\0') {
	err_code = ENOENT;
	return(NIL_INODE);
  }

  dup_inode(rip);		/* inode will be returned with put_inode */
  flog(LOG_DEBUG,"i_nlinks:%d path:%s i_num:%d",rip->i_nlinks,path,rip->i_num);

  /* Scan the path component by component. */
  while (TRUE) {
	/* Extract one component. */
	if ( (new_name = get_name(path, string)) == (char*) 0) {
		put_inode(rip);	/* bad path in user space */
		return(NIL_INODE);
	}

	//flog(LOG_DEBUG,"new_name %s",new_name);
	if (*new_name == '\0') {
		if ( (rip->i_mode & I_TYPE) == I_DIRECTORY) {
			return(rip);	/* normal exit */
		} else {
			/* last file of path prefix is not a directory */
			put_inode(rip);
			err_code = ENOTDIR;			
			return(NIL_INODE);
		}
    }

	/* There is more path.  Keep parsing. */
	new_ip = advance(rip, string);
	put_inode(rip);		/* rip either obsolete or irrelevant */
	if (new_ip == NIL_INODE) return(NIL_INODE);

	/* The call to advance() succeeded.  Fetch next component. */
	path = new_name;
	rip = new_ip;
  }
}

/*===========================================================================*
 *				get_name				     *
 *===========================================================================*/
//char *old_name;			/* path name to parse */
//char string[NAME_MAX];		/* component extracted from 'old_name' */

PRIVATE
char *get_name(const char *old_name, char string[NAME_MAX])
{
/* Given a pointer to a path name in fs space, 'old_name', copy the next
 * component to 'string' and pad with zeros.  A pointer to that part of
 * the name as yet unparsed is returned.  Roughly speaking,
 * 'get_name' = 'old_name' - 'string'.
 *
 * This routine follows the standard convention that /usr/ast, /usr//ast,
 * //usr///ast and /usr/ast/ are all equivalent.
 */

	int c;
	char *np;
	const char *rnp;

	np = string;			/* 'np' points to current position */
	rnp = old_name;		/* 'rnp' points to unparsed string */
	while ( (c = *rnp) == '/') rnp++;	/* skip leading slashes */

  /* Copy the unparsed path, 'old_name', to the array, 'string'. */
	while ( rnp < &old_name[PATH_MAX]  &&  c != '/' &&  c != '\0') {
		if (np < &string[NAME_MAX]) *np++ = c;
		c = *++rnp;		/* advance to next character */
	}

	/* To make /usr/ast/ equivalent to /usr/ast, skip trailing slashes. */
	while (c == '/' && rnp < &old_name[PATH_MAX]) c = *++rnp;

	if (np < &string[NAME_MAX]) *np = '\0';	/* Terminate string */

	if (rnp >= &old_name[PATH_MAX]) {
		err_code = ENAMETOOLONG;
		return((char *) 0);
	}
	return(const_cast<char*>(rnp));
}

/*===========================================================================*
 *				advance					     *
 *===========================================================================*/
//struct inode *dirp;		/* inode for directory to be searched */
//char string[NAME_MAX];		/* component name to look for */
 PUBLIC
inode *advance(inode *dirp, char string[NAME_MAX])
{
/* Given a directory and a component of a path, look up the component in
 * the directory, find the inode, open it, and return a pointer to its inode
 * slot.  If it can't be done, return NIL_INODE.
 */

  inode *rip;
  inode *rip2;
  super_block *sp;
  int r, inumb;
  dev_t mnt_dev;
  ino_t numb;
  fproc *fp = get_fproc(getpid()); //questo kernel è monolitico

  //flog(LOG_DEBUG,"advance string:%s",string);
  /* If 'string' is empty, yield same inode straight away. */
  if (string[0] == '\0') { return(get_inode(dirp->i_dev, (int) dirp->i_num)); }

  /* Check for NIL_INODE. */
  if (dirp == NIL_INODE) { return(NIL_INODE); }

  /* If 'string' is not present in the directory, signal error. */
  if ( (r = search_dir(dirp, string, &numb, LOOK_UP)) != OK) {
	flog(LOG_DEBUG,"errore search_dir");
	err_code = r;
	return(NIL_INODE);
  }

  /* Don't go beyond the current root directory, unless the string is dot2. */
  if (dirp == fp->fp_rootdir && strcmp(string, "..") == 0 && string != dot2)
		return(get_inode(dirp->i_dev, (int) dirp->i_num));

  /* The component has been found in the directory.  Get inode. */
  if ( (rip = get_inode(dirp->i_dev, (int) numb)) == NIL_INODE)  {
	return(NIL_INODE);
	}

  if (rip->i_num == ROOT_INODE)
	if (dirp->i_num == ROOT_INODE) {
	    if (string[1] == '.') {
		for (sp = &super_b[1]; sp < &super_b[NR_SUPERS]; sp++){
				if (sp->s_dev == rip->i_dev) {
					/* Release the root inode.  Replace by the
					 * inode mounted on.
					 */
					put_inode(rip);
					mnt_dev = sp->s_imount->i_dev;
					inumb = (int) sp->s_imount->i_num;
					rip2 = get_inode(mnt_dev, inumb);
					rip = advance(rip2, string);
					put_inode(rip2);
					break;
				}
			}
		}
	}
  if (rip == NIL_INODE) return(NIL_INODE);

  /* See if the inode is mounted on.  If so, switch to root directory of the
   * mounted file system.  The super_block provides the linkage between the
   * inode mounted on and the root directory of the mounted file system.
   */
  while (rip != NIL_INODE && rip->i_mount == I_MOUNT) {
	/* The inode is indeed mounted on. */
	for (sp = &super_b[0]; sp < &super_b[NR_SUPERS]; sp++) {
		if (sp->s_imount == rip) {
			/* Release the inode mounted on.  Replace by the
			 * inode of the root inode of the mounted device.
			 */
			put_inode(rip);
			rip = get_inode(sp->s_dev, ROOT_INODE);
			break;
		}
	}
  }
  return(rip);		/* return pointer to inode's component */
}



/*===========================================================================*
 *				search_dir				     *
 * modificata per supportare minix V1 e V2
 *===========================================================================*/
//register struct inode *ldir_ptr; /* ptr to inode for dir to search */
//char string[NAME_MAX];		 /* component to search for */
//ino_t *numb;			 /* pointer to inode number */
//int flag;			 /* LOOK_UP, ENTER, DELETE or IS_EMPTY */
PUBLIC int search_dir(inode *ldir_ptr, char *string, ino_t *numb, int flag)
{
/* This function searches the directory whose inode is pointed to by 'ldip':
 * if (flag == ENTER)  enter 'string' in the directory with inode # '*numb';
 * if (flag == DELETE) delete 'string' from the directory;
 * if (flag == LOOK_UP) search for 'string' and return inode # in 'numb';
 * if (flag == IS_EMPTY) return OK if only . and .. in dir else ENOTEMPTY;
 *
 *    if 'string' is dot1 or dot2, no access permissions are checked.
 */

	//direct *dp = NULL;
	//char** dp_byte = (char**)&dp;

 // In questa funzione si usano dei trucchi di programmazione 
 // sporchi per gestire i vari casi possibili
 // L'inode number può essere grande 2B oppure 4B
 // Il nome del file può essere lungo 14B, 32B oppure 60B
 // Non è possibile usare un puntatore direct* per
 // per tutti i casi possibili, quindi per leggere e modificare
 // la cartella uso un semplice puntatore a carattere + offset(variabili).
 	char *dp_byte = NULL; 

	//struct buf *bp = NULL;
	blocco_fis *bp;
	int i, r, e_hit, t, match;
	mode_t bits;
	off_t pos;
	unsigned new_slots, old_slots;
	block_t b;
	super_block *sp = ldir_ptr->i_sp;
	int extended = 0;
	char version = sp->s_disk_version;
	uint8_t dim_dir; // dimensione di una cartella
	uint8_t dim_name; //numero di byte dedicati al nome
	uint8_t dim_ino_numb; //numero di byte dedicati all'inode numb;

	if( version == V1 || version == V2 ){
		dim_dir = 16;
		dim_name = 14;
		dim_ino_numb = 2;
	} else if ( version == V1_REV || version == V2_REV ) {
		dim_dir = 32;
		dim_name = 30;
		dim_ino_numb = 2;
	} else {
		dim_dir = 64;
		dim_name = 60;
		dim_ino_numb = 4;
	}

  /* If 'ldir_ptr' is not a pointer to a dir inode, error. */
	if ( (ldir_ptr->i_mode & I_TYPE) != I_DIRECTORY){
   		flog(LOG_DEBUG,"ldir_ptr is not a pointer to a dir inode");
		return(ENOTDIR);
  	}
  	r = OK;

  	if (flag != IS_EMPTY) {
		bits = (flag == LOOK_UP ? X_BIT : W_BIT | X_BIT);

		if (string == dot1 || string == dot2) {
			if (flag != LOOK_UP) r = read_only(ldir_ptr);
					     /* only a writable device is required. */
		}
		else r = forbidden(ldir_ptr, bits); /* check access permissions */
  	}
  	if (r != OK) return(r);
  
	// Step through the directory one block at a time. 
	old_slots = (unsigned) (ldir_ptr->i_size/dim_dir);
	new_slots = 0;
	e_hit = FALSE;
	match = 0;			// set when a string match occurs 

    //flog(LOG_DEBUG,"scansione cartella");
	for (pos = 0; pos < ldir_ptr->i_size; pos += ldir_ptr->i_sp->s_block_size) {
		b = read_map(ldir_ptr, pos);	/* get block number */
		//flog(LOG_DEBUG,"cartella- i_num:%d pos:%d blocco %d",ldir_ptr->i_num,pos,b);
		// Since directories don't have holes, 'b' cannot be NO_BLOCK. 
		bp = get_block(ldir_ptr->i_dev, b, NORMAL);	// get a dir block 
		if (bp == NO_BLOCK)
			panic("get_block returned NO_BLOCK");

		// Search a directory block. 
		/*for (dp = &bp->b_dir[0];
			 dp < &bp->b_dir[NR_DIR_ENTRIES(ldir_ptr->i_sp->s_block_size)];
			 dp++) */
		//flog(LOG_DEBUG,"dim_dir=%d",dim_dir);
		for (dp_byte = (char*)bp;
			 dp_byte < (char*)bp + NR_DIR_ENTRIES(ldir_ptr->i_sp->s_block_size,dim_dir)*dim_dir;
			 dp_byte += dim_dir ) 
		{
			flog(LOG_DEBUG,"dp_byte:%p",dp_byte);
			//flog(LOG_DEBUG,"dp->d_ino:%d dp->d_name:%s NR_DIR_ENTRIES:%d",*((short*)dp_byte),dp_byte+dim_ino_numb,NR_DIR_ENTRIES(ldir_ptr->i_sp->s_block_size,dim_dir));
			if (++new_slots > old_slots) { /* not found, but room left */
				if (flag == ENTER) e_hit = TRUE;
				break;
			}

			// Match occurs if string found. 
			if (flag != ENTER && *((short*)dp_byte) != 0) {
				if (flag == IS_EMPTY) {
					/* If this test succeeds, dir is not empty. */
					if (strcmp(dp_byte + dim_ino_numb, "." ) != 0 && strcmp(dp_byte + dim_ino_numb, "..") != 0) 
					    match = 1;
				} else {
					//flog(LOG_DEBUG,"search_dir dp->d_name:%s string:%s strncmp:%d strcmp:%d",dp->d_name,string,strncmp(dp->d_name, string, NAME_MAX),strcmp(dp->d_name,string));
					if (strncmp(dp_byte + dim_ino_numb , string, dim_name) == 0) {
						match = 1;
					}
				}
			}

			if (match) {
				/* LOOK_UP or DELETE found what it wanted. */
				r = OK;
				if (flag == IS_EMPTY) r = ENOTEMPTY;
				else if (flag == DELETE) {
					// Save d_ino for recovery. 
					t = dim_name - dim_ino_numb;
					//*((ino_t *) &dp->d_name[t]) = dp->d_ino;
					memcpy(dp_byte + t, dp_byte, dim_ino_numb); // salva l'inode number per recovery
					memset(dp_byte, 0, dim_ino_numb); //imposto l'inode number a 0
					//dp->d_ino = 0;	// erase entry 
					//bp->b_dirt = DIRTY;
					ldir_ptr->i_update |= CTIME | MTIME;
					ldir_ptr->i_dirt = DIRTY;
				} else {
					sp = ldir_ptr->i_sp;	/* 'flag' is LOOK_UP */
					if( dim_ino_numb == 4 )
						*numb = conv4(sp->s_native, (int) ((direct*)dp_byte)->d_ino);
					else
						*numb = conv2(sp->s_native, (short) ((direct_V1*)dp_byte)->d_ino);
				}
				//put_block(bp, DIRECTORY_BLOCK);
				//c_dmawritehd_n((uint16_t*)bp,b,1,errore);
				flog(LOG_DEBUG, "sp:%p sp->s_dev:%d",sp,sp->s_dev);
				put_block(sp->s_dev,b,bp);
				return(r);
			}

			// Check for free slot for the benefit of ENTER. 
			// dp->d_ino == 0
			if (flag == ENTER && ((short*)dp_byte)[0] == 0) {
				e_hit = TRUE;	// we found a free slot 
				break;
			}
		}

		// The whole block has been searched or ENTER has a free slot. 
		if (e_hit) break;	// e_hit set if ENTER can be performed now 
		//put_block(bp, DIRECTORY_BLOCK);	// otherwise, continue searching dir 
		put_block(ldir_ptr->i_dev,b,bp);
  	}

	/* The whole directory has now been searched. */
	if (flag != ENTER) {
		return(flag == IS_EMPTY ? OK : ENOENT);
	}

  /* This call is for ENTER.  If no free slot has been found so far, try to
   * extend directory.
   */
	if (e_hit == FALSE) { /* directory is full and no room left in last block */
		new_slots++;		/* increase directory size by 1 entry */
		if (new_slots == 0) return(EFBIG); /* dir size limited by slot count */
		if ( (bp = new_block(ldir_ptr, ldir_ptr->i_size)) == /*NIL_BUF*/0)
			return(err_code);
		dp_byte = (char*)&bp->b_dir[0];
		extended = 1;
  	}

	/* 'bp' now points to a directory block with space. 'dp' points to slot. */
	memset(dp_byte + dim_ino_numb, 0, dim_name); // pulisco l'entrata
	for (i = 0; string[i] && i < dim_name; i++)
		*(dp_byte + dim_ino_numb + i) = string[i]; //copio la stringa
	sp = ldir_ptr->i_sp; 
	if( dim_ino_numb == 4 )
		((direct*)dp_byte)->d_ino = conv4(sp->s_native,(int) *numb);
	else
		((direct_V1*)dp_byte)->d_ino = conv2(sp->s_native,(short) *numb);
	//bp->b_dirt = DIRTY;
	//put_block(bp, DIRECTORY_BLOCK);
	put_block(ldir_ptr->i_dev,b,bp);
	ldir_ptr->i_update |= CTIME | MTIME;	/* mark mtime for update later */
	ldir_ptr->i_dirt = DIRTY;
	if (new_slots > old_slots) {
		ldir_ptr->i_size = (off_t) new_slots * dim_dir;
		/* Send the change to disk if the directory is extended. */
		if (extended) rw_inode(ldir_ptr, WRITING);
	}
	return(OK);
}
