/* This file contains the heart of the mechanism used to read (and write)
 * files.  Read and write requests are split up into chunks that do not cross
 * block boundaries.  Each chunk is then processed in turn.  Reads on special
 * files are also detected and handled.
 *
 * The entry points into this file are
 *   do_read:	 perform the READ system call by calling read_write
 *   read_write: actually do the work of READ and WRITE
 *   read_map:	 given an inode and file position, look up its zone number
 *   rd_indir:	 read an entry in an indirect block 
 *   read_ahead: manage the block read ahead business
 */
#include "libce.h"
#include "fs.h"
#include "sys/fcntl.h"
//#include <minix/com.h>
//#include "buf.h"
#include "file.h"
#include "fproc.h"
#include "inode.h"
//#include "param.h"
#include "glo.h"


inode *rdahed_inode;
off_t rdahedpos;         /* position to read ahead */

PRIVATE int rw_chunk(struct inode *rip, off_t position,
	unsigned off, int chunk, unsigned left, int rw_flag,
	char *buff, int block_size, int *completed);

/*===========================================================================*
 *				do_read					     *
 *===========================================================================*/
PUBLIC 
int do_read(int fd, void *buf, uint32_t nbytes)
{
	//flog(LOG_DEBUG,"do_read %d nbytes:%d",buf,nbytes);
	// la memoria allocata nello heap è residente 
	// ed inoltre è nello spazio condiviso
	void *local_buf = mem_alloc(nbytes);
	memset(local_buf,0,nbytes);
	int res = read_write(READING,fd,static_cast<char*>(local_buf),nbytes);
	memcpy(buf,local_buf,nbytes);
	mem_free(local_buf);
	return res;
}

/*===========================================================================*
 *				read_write				     *
 * modificata rispetto a quella originale visto che il nostro kernel 
 * è monolitico
 *===========================================================================*/
PUBLIC int read_write(int rw_flag, int fd, char* buff, uint32_t nbytes)
//int rw_flag;			/* READING or WRITING */
{
// Perform read(fd, buffer, nbytes) or write(fd, buffer, nbytes) call. 
	inode *rip;
	filp *f;
	off_t bytes_left, f_size, position;
	unsigned int off, cum_io=0;
	int oflags, r, chunk, block_spec, char_spec;
	int regular/*, partial_pipe = 0, partial_cnt = 0*/;
	mode_t mode_word;
	int block_size;
	int completed;

  	// If the file descriptor is valid, get the inode, size and mode. 
	if (nbytes < 0) return(EINVAL);
	if ((f = get_filp(fd)) == NIL_FILP){
		return(err_code);
	}

	// mutua esclusione sul file
	sem_wait(f->mutex);

	//flog(LOG_DEBUG,"filp_mode:%x R_BIT:%x W_BIT:%x",f->filp_mode,R_BIT,W_BIT);
	if (((f->filp_mode) & (rw_flag == READING ? R_BIT : W_BIT)) == 0) {
		sem_signal(f->mutex);
		return(f->filp_mode == FILP_CLOSED ? EIO : EBADF);
	}
	if (nbytes == 0) {
		 return(0);	/* so char special files need not check for 0*/
		sem_signal(f->mutex);
	}

	position = f->filp_pos;
	oflags = f->filp_flags;
	rip = f->filp_ino;
	f_size = rip->i_size;
	r = OK;

	mode_word = rip->i_mode & I_TYPE;
	regular = mode_word == I_REGULAR || mode_word == I_NAMED_PIPE;

	if ((char_spec = (mode_word == I_CHAR_SPECIAL ? 1 : 0))) {
		if (rip->i_zone[0] == NO_DEV)
			panic("read_write tries to read from character device NO_DEV");
		block_size = get_block_size(rip->i_zone[0]);
	}
  	if ((block_spec = (mode_word == I_BLOCK_SPECIAL ? 1 : 0))) {
		f_size = UINT_MAX;
		if (rip->i_zone[0] == NO_DEV)
			panic("read_write tries to read from block device NO_DEV");
		block_size = get_block_size(rip->i_zone[0]);
  	}

  	if (!char_spec && !block_spec)
  		block_size = rip->i_sp->s_block_size;


  /* Check for character special files. */
  	if (char_spec) {
	  	/*dev_t dev;
		dev = (dev_t) rip->i_zone[0];
		r = dev_io(op, dev, usr, m_in.buffer, position, m_in.nbytes, oflags);
		if( rw_flag == READING ){	
			read_bytes_hd(position, (char*)buff, nbytes);
			flog(LOG_DEBUG,"char_spec READING");
		} else {
			flog(LOG_DEBUG,"char_spec WRITING");
		}

		if (r >= 0) {
			cum_io = r;
			position += r;
			r = OK;
		}*/
		panic("char_spec");
  	} else { 
		if (rw_flag == WRITING && block_spec == 0) {
			// Check in advance to see if file will grow too big. 
			if (position > rip->i_sp->s_max_size - nbytes) {
				sem_signal(f->mutex);
				return(EFBIG);
			}

			// Check for O_APPEND flag. 
			if (oflags & O_APPEND) position = f_size;

			/* Clear the zone containing present EOF if hole about
			 * to be created.  This is necessary because all unwritten
			 * blocks prior to the EOF must read as zeros.
			 */
			if (position > f_size) clear_zone(rip, f_size, 0);
		}

		// Split the transfer into chunks that don't span two blocks. 
		while (nbytes != 0) {

			off = (unsigned int) (position % block_size);// offset in blk
			chunk = MIN(nbytes, block_size - off);
			//flog(LOG_DEBUG,"read_write nbytes:%d chunk:%d block_size:%d off:%d f_size:%d",nbytes,chunk, block_size,off,f_size);
			if (chunk < 0) chunk = block_size - off;

			if (rw_flag == READING) {
				bytes_left = f_size - position;
				if (position >= f_size) break;	/* we are beyond EOF */
				if (chunk > (int)bytes_left) chunk = (int) bytes_left;
			}

			/* Read or write 'chunk' bytes. */
			r = rw_chunk(rip, position, off, chunk, (unsigned) nbytes,
				     rw_flag, (char*)buff, block_size, &completed);

			if (r != OK) break;	/* EOF reached */
			if (rdwt_err < 0) break;

			/* Update counters and pointers. */
			buff += chunk;	/* user buffer address */
			nbytes -= chunk;	/* bytes yet to be read */
			cum_io += chunk;	/* bytes read so far */
			position += chunk;	/* position within the file */

		}
  	}

  /* On write, update file size and access time. */
	if (rw_flag == WRITING) {
		if (regular || mode_word == I_DIRECTORY) {
			if (position > f_size) rip->i_size = position;
		}
	} 
	flog(LOG_DEBUG, "read_write f:%p f->filp_pos:%d position:%d rip->i_size:%d",f,f->filp_pos,position,rip->i_size);
  	f->filp_pos = position;

	/* Check to see if read-ahead is called for, and if so, set it up. */
	if (rw_flag == READING && rip->i_seek == NO_SEEK && position % block_size== 0
		&& (regular || mode_word == I_DIRECTORY)) {
		rdahed_inode = rip;
		rdahedpos = position;
	}
	rip->i_seek = NO_SEEK;

	if (rdwt_err != OK) r = rdwt_err;	// check for disk error
	if (rdwt_err == END_OF_FILE) r = OK;

	if (r == OK) {
		if (rw_flag == READING) rip->i_update |= ATIME;
		if (rw_flag == WRITING) rip->i_update |= CTIME | MTIME;
		rip->i_dirt = DIRTY;		/* inode is thus now dirty */
		sem_signal(f->mutex);
		return(cum_io);
  	}

  	sem_signal(f->mutex);
	return(r);
}

/*===========================================================================*
 *				rw_chunk				     *
 *===========================================================================*/
PRIVATE int rw_chunk(inode *rip, off_t position, unsigned off, int chunk, 
                     unsigned left, int rw_flag, char *buff,
                     int block_size, int *completed)
//register struct inode *rip;	/* pointer to inode for file to be rd/wr */
//off_t position;			/* position within file to read or write */
//unsigned off;			/* off within the current block */
//int chunk;			/* number of bytes to read or write */
//unsigned left;			/* max number of bytes wanted after position */
//int rw_flag;			/* READING or WRITING */
//char *buff;			/* virtual address of the user buffer */
//int seg;			/* T or D segment in user space */
//int usr;			/* which user process */
//int block_size;			/* block size of FS operating on */
//int *completed;			/* number of bytes copied */
{
/* Read or write (part of) a block. */

	blocco_fis *bp;
	int r = OK;
	int n, block_spec;
	block_t b;
	dev_t dev;

	*completed = 0;
	block_spec = (rip->i_mode & I_TYPE) == I_BLOCK_SPECIAL;
  	if (block_spec) {
		b = position/block_size;
		dev = (dev_t) rip->i_zone[0];
  	} else {
		b = read_map(rip, position);
		dev = rip->i_dev;
  	}
  	flog(LOG_DEBUG,"!rw_chunk i_num:%d chunk:%d b=%d block_size:%d position:%d size:%d",rip->i_num,chunk,b,block_size,position,rip->i_size);

  	if (!block_spec && b == NO_BLOCK) {
		if (rw_flag == READING) {
			// Reading from a nonexistent block.  Must read as all zeros.
			bp = get_block(NO_DEV, NO_BLOCK, NORMAL);    // get a buffer 
			zero_block(bp);
		} else {
			//flog(LOG_DEBUG,"rw_chunk - creo un nuovo blocco per l'inode %d",rip->i_num);
			// Writing to a nonexistent block. Create and enter in inode.
			if ((bp = new_block(rip, position)) == NIL_BLK)return(err_code);
		}
  	} else if (rw_flag == READING) {
		/* Read and read ahead if convenient. */
		//bp = rahead(rip, b, position, left);
		bp = bloccofis_file(rip, position);
  	} else {
		/* Normally an existing block to be partially overwritten is first read
		 * in.  However, a full block need not be read in.  If it is already in
		 * the cache, acquire it, otherwise just acquire a free buffer.
		 */
		/* //NB:la cache non esiste nella mia implementazione
		n = (chunk == block_size ? NO_READ : NORMAL);
		if (!block_spec && off == 0 && position >= rip->i_size) n = NO_READ; */
		n = NORMAL;
		bp = get_block(dev, b, n);
  	}

  	/* In all cases, bp now points to a valid buffer. */
  	if (bp == NIL_BLK) {
  		panic("bp not valid in rw_chunk, this can't happen");
  	}
  	if (rw_flag == WRITING && chunk != block_size && !block_spec && position >= rip->i_size && off == 0) {
  		//flog(LOG_DEBUG,"rw_chunk - azzero blocco");
		zero_block(bp);
  	}

 /* if (rw_flag == READING) {
	// Copy a chunk from the block buffer to user space. 
	r = sys_vircopy(FS_PROC_NR, D, (phys_bytes) (bp->b_data+off),
			usr, seg, (phys_bytes) buff,
			(phys_bytes) chunk);
  } else {
	// Copy a chunk from user space to the block buffer.
	r = sys_vircopy(usr, seg, (phys_bytes) buff,
			FS_PROC_NR, D, (phys_bytes) (bp->b_data+off),
			(phys_bytes) chunk);
	bp->b_dirt = DIRTY;
  }*/
	if( rw_flag == READING ) {
		//flog(LOG_DEBUG, "!!!rw_chunk READING chunk:%d off:%d buff:%x",chunk,off,buff);
		memcpy(buff,bp->b_data+off,chunk);
		r = OK;
	} else {
		memcpy(bp->b_data+off,buff,chunk);
		r = OK;
		b = read_map(rip,position);
	//	flog(LOG_DEBUG,"!!!rw_chunk - scrivo il blocco in posizione %d",b);
		put_block(dev, b, bp);
		//c_dmawritehd_n((natw*)bp,b,1,errore);
	}
  	//n = (off + chunk == block_size ? FULL_DATA_BLOCK : PARTIAL_DATA_BLOCK);
	//put_block(bp, n);
	mem_free(bp);
  	return(r);
}


/*===========================================================================*
 *				read_map				     *
 *===========================================================================*/
PUBLIC block_t read_map(inode *rip, off_t position)
//register struct inode *rip;	/* ptr to inode to map from */
//off_t position;			/* position in file whose blk wanted */
{
/* Given an inode and a position within the corresponding file, locate the
 * block (not zone) number in which that position is to be found and return it.
 */

  //register struct buf *bp;
 	blocco_fis *bp;
	zone_t z;
	int scale, boff, dzones, nr_indirects, index, zind, ex;
	block_t b;
	long excess, zone, block_pos;
  
	scale = rip->i_sp->s_log_zone_size;	// for block-zone conversion 
	block_pos = position/rip->i_sp->s_block_size;	// relative blk # in file 
	zone = block_pos >> scale;	// position's zone 
	boff = (int) (block_pos - (zone << scale) ); // relative blk # within zone 
	dzones = rip->i_ndzones;
  	nr_indirects = rip->i_nindirs;
  
  	// Is 'position' to be found in the inode itself? 
  	if (zone < dzones) {
		zind = (int) zone;	// index should be an int 
		z = rip->i_zone[zind];
		if (z == NO_ZONE) return(NO_BLOCK);
		b = ((block_t) z << scale) + boff;
		return b;
	}

  	// It is not in the inode, so it must be single or double indirect. 
  	excess = zone - dzones;	// first Vx_NR_DZONES don't count 

  	if (excess < nr_indirects) {
		// 'position' can be located via the single indirect block. 
		z = rip->i_zone[dzones];
  	} else {
		// 'position' can be located via the double indirect block. 
		if ( (z = rip->i_zone[dzones+1]) == NO_ZONE) return(NO_BLOCK);
		excess -= nr_indirects;			// single indir doesn't count
		b = (block_t) z << scale;
		bp = get_block(rip->i_dev, b, NORMAL);	// get double indirect block 
		index = (int) (excess/nr_indirects);
		z = rd_indir(bp, index);		// z= zone for single
		//put_block(bp, INDIRECT_BLOCK);		// release double ind block 
		put_block(rip->i_dev, b, bp);
		excess = excess % nr_indirects;		// index into single ind blk 
	}

	// 'z' is zone num for single indirect block; 'excess' is index into it. 
	if (z == NO_ZONE) return(NO_BLOCK);
	b = (block_t) z << scale;			// b is blk # for single ind 
	bp = get_block(rip->i_dev, b, NORMAL);	// get single indirect block 
	ex = (int) excess;				// need an integer 
	z = rd_indir(bp, ex);			// get block pointed to 
	//put_block(bp, INDIRECT_BLOCK); // release single indir blk
	put_block(rip->i_dev,b,bp);
	if (z == NO_ZONE) return(NO_BLOCK);
	b = ((block_t) z << scale) + boff;
	return(b);
}

/*===========================================================================*
 *				rd_indir				     *
 *===========================================================================*/
PUBLIC zone_t rd_indir(blocco_fis *bp, int index)
//struct buf *bp;			/* pointer to indirect block */
//int index;			/* index into *bp */
{
	/* Given a pointer to an indirect block, read one entry.  The reason for
	* making a separate routine out of this is that there are four cases:
	* V1 (IBM and 68000), and V2 (IBM and 68000).
	*/

	struct super_block *sp;
	zone_t zone;			/* V2 zones are longs (shorts in V1) */

	sp = get_super(CE_DEV);	/* need super block to find file sys type */

	/* read a zone from an indirect block */
	if (sp->s_version == V1)
		zone = (zone_t) conv2(sp->s_native, (int)  bp->b_v1_ind[index]);
	else
		zone = (zone_t) conv4(sp->s_native, (long) bp->b_v2_ind[index]);

	if (zone != NO_ZONE && (zone < (zone_t) sp->s_firstdatazone || zone >= sp->s_zones)) {
		flog(LOG_ERR,"Illegal zone number %ld in indirect block, index %d\n",zone, index);
		panic("check file system");
	}
	return(zone);
}
