/* This file manages the super block table and the related data structures,
 * namely, the bit maps that keep track of which zones and which inodes are
 * allocated and which are free.  When a new inode or zone is needed, the
 * appropriate bit map is searched for a free entry.
 *
 * The entry points into this file are
 *   alloc_bit:       somebody wants to allocate a zone or inode; find one
 *   free_bit:        indicate that a zone or inode is available for allocation
 *   get_super:       search the 'superblock' table for a device
 *   mounted:         tells if file inode is on mounted (or ROOT) file system
 *   read_super:      read a superblock
 */

#include "libce.h"
#include "fs.h"
#include "glo.h"
//#include <string.h>
//#include <minix/com.h>
//#include "buf.h"
#include "inode.h"
#include "super.h"
#include "const.h"

char *inode_map = NULL;
char *zone_map = NULL;

extern "C" void c_dmareadhd_n(uint16_t vetti[], uint32_t primo,
		uint16_t quanti, uint8_t &errore);

/*===========================================================================*
 *				alloc_bit				     *
 *===========================================================================*/
PUBLIC bit_t alloc_bit(super_block *sp, int map, bit_t origin)
//struct super_block *sp;		/* the filesystem to allocate from */
//int map;			/* IMAP (inode map) or ZMAP (zone map) */
//bit_t origin;			/* number of bit to start searching at */
{
	bit_t i = origin/8;
	bit_t j = origin%8;
	//uint8_t mask = 0x80;
	uint8_t mask = 0x01;
	bit_t n_bits; //numero di bit nella bitmap
	char *bitmap;

	if (map == IMAP) {
		bitmap = inode_map;
		n_bits = sp->s_zmap_blocks * sp->s_block_size * 8;
	} else {
		bitmap = zone_map;
		n_bits = sp->s_zmap_blocks * sp->s_block_size*8;
	}

	for( i=0; i<n_bits/8; i+=j)
	{
		flog(LOG_DEBUG,"bitmap[%d]=%x mask=%x",i,bitmap[i],mask);
		for(j=0; j<8 && i+j<n_bits; j++){
			if( (bitmap[i] & (mask << j)) == 0 ){
				bitmap[i] |= (mask << j);

			   //poco ottimizzato ma semplicissimo
				if( map == IMAP )
					_write_imap();
			   	else
					_write_zmap(); 
			   	flog(LOG_DEBUG,"map:%d alloc_bit n. %d",bitmap[i],i+j);
			   	return i + j;
			}
		}
	}

	return(NO_BIT);		// no bit could be allocated 
}
/*===========================================================================*
 *				free_bit				     *
 *===========================================================================*/
PUBLIC void free_bit(super_block *sp, int map, bit_t bit_returned)
//struct super_block *sp;		/* the filesystem to operate on */
//int map;			/* IMAP (inode map) or ZMAP (zone map) */
//bit_t bit_returned;		/* number of bit to insert into the map */
{
/* Return a zone or inode by turning off its bitmap bit. */

	bit_t i = bit_returned/8;
	bit_t j = bit_returned%8;
	uint8_t mask = 0x01;
	char* bitmap;

	if ( map == IMAP ) {
		bitmap = inode_map;
	} else {
		bitmap = zone_map;
	}

	bitmap[i] &= ~(mask << j);

	//poco ottimizzato ma semplicissimo
	if( map == IMAP )
		_write_imap();
	else
		_write_zmap();
}

/*===========================================================================*
 *				get_super				     *
 *===========================================================================*/
PUBLIC struct super_block *get_super(dev_t dev)
//dev_t dev;			/* device number whose super_block is sought */
{
/* Search the superblock table for this device.  It is supposed to be there. */

	super_block *sp;

	if (dev == NO_DEV)
		panic("request for super_block of NO_DEV");

	for (sp = &super_b[0]; sp < &super_b[NR_SUPERS]; sp++){
		//flog(LOG_DEBUG,"get_super sp=%x",sp);
		if (sp->s_dev == dev) return(sp);
	}

	/* Search failed.  Something wrong. */
	panic("can't find superblock for device (in decimal)");

	return(NIL_SUPER);		/* to keep the compiler and lint quiet */
}

/*===========================================================================*
 *				get_block_size				     *
 *===========================================================================*/
PUBLIC int get_block_size(dev_t dev)
{
/* Search the superblock table for this device. */

  register struct super_block *sp;

  if (dev == NO_DEV)
  	panic("request for block size of NO_DEV");

  for (sp = &super_b[0]; sp < &super_b[NR_SUPERS]; sp++) {
  	if (sp->s_dev == dev) {
  		return(sp->s_block_size);
  	}
  }

  /* no mounted filesystem? use this block size then. */
  return BLOCK_SIZE;
}

/*===========================================================================*
 *				mounted					     *
 *===========================================================================*/
PUBLIC int mounted(inode *rip)
//register struct inode *rip;	/* pointer to inode */
{
/* Report on whether the given inode is on a mounted (or ROOT) file system. */

	super_block *sp;
	dev_t dev = CE_DEV; //l'unico disco

	//dev = (dev_t) rip->i_zone[0];
	//if (dev == root_dev) return(TRUE);	/* inode is on root file system */

	for (sp = &super_b[0]; sp < &super_b[NR_SUPERS]; sp++)
		if (sp->s_dev == dev) return(TRUE);

	return(FALSE);
}

void trasforma_superblocco(minix3_super_block *out, minix_super_block *in)
{
	out->s_ninodes = in->s_ninodes;
	out->s_nzones = in->s_nzones;
	out->s_imap_blocks = in->s_imap_blocks;
	out->s_zmap_blocks = in->s_zmap_blocks;
	out->s_firstdatazone = in->s_firstdatazone;
	out->s_log_zone_size = in->s_log_zone_size; 
	out->s_max_size = in->s_max_size; 
	out->s_magic = in->s_imap_blocks;
	out->s_zones = in->s_zones;
}

/*===========================================================================*
 *        read_super             *
 * versione adattata per funzionare con il nucleo
 *===========================================================================*/
PUBLIC int read_super(super_block *sp)
//register struct super_block *sp; /* pointer to a superblock */
{
// Read a superblock. 
	dev_t dev;
	int magic;
	int version, native;
	uint8_t errore;
	char *buf = (char*)mem_alloc(SUPER_BLOCK_BYTES);

	if( !buf )
		panic("heap io pieno.");

	dev = sp->s_dev;    /* save device (will be overwritten by copy) */
	if (dev == NO_DEV)
		panic("request for super_block of NO_DEV");
	c_dmareadhd_n((natw*)buf, BOOT_BLOCK_BYTES/DIM_SECTOR, SUPER_BLOCK_BYTES/DIM_SECTOR, errore);

	sp->s_dev = NO_DEV;   /* restore later */
	// devo riconoscere la versione usata leggendo il magic number
	// L'offset del magic number però non è costante con le varie
	// versioni usate quindi devo testare tutti i possibili casi
	magic = ((minix_super_block*)buf)->s_magic;
	if( magic == MINIX_SUPER_MAGIC || magic == conv2(BYTE_SWAP, MINIX_SUPER_MAGIC) )
	{
		flog(LOG_DEBUG, "versione 1 rev1");
		// nomi di max 14 caratteri
		version = V1;
		native = (magic == MINIX_SUPER_MAGIC);
		trasforma_superblocco((minix3_super_block*)sp, (minix_super_block*)buf);	
	} else if ( magic == MINIX_SUPER_MAGIC2 || magic == conv2(BYTE_SWAP, MINIX_SUPER_MAGIC2)) {
		flog(LOG_DEBUG, "versione 1 rev2");
		// nomi di max 30 caratteri
		version = V1_REV;
		native = (magic == MINIX_SUPER_MAGIC2);
		trasforma_superblocco((minix3_super_block*)sp, (minix_super_block*)buf);		
	} else if( magic == MINIX2_SUPER_MAGIC || magic == conv2(BYTE_SWAP, MINIX2_SUPER_MAGIC) ){
		flog(LOG_DEBUG, "versione 2 rev1");
		version = V2;
		// nomi di max 14 caratteri
		native = (magic == MINIX2_SUPER_MAGIC);
		trasforma_superblocco((minix3_super_block*)sp, (minix_super_block*)buf);
	} else if( magic == MINIX2_SUPER_MAGIC2  || magic == conv2(BYTE_SWAP, MINIX2_SUPER_MAGIC2) ) {
		flog(LOG_DEBUG, "versione 2 rev2");
		version = V2_REV;
		// nomi di max 30 caratteri
		native  = (magic == MINIX2_SUPER_MAGIC2);
		trasforma_superblocco((minix3_super_block*)sp, (minix_super_block*)buf);
	} else {
		// provo se il formato è quello di minxi v3
		magic = ((minix3_super_block*)buf)->s_magic;
		if ( magic == MINIX3_SUPER_MAGIC ) {
			// il formato del superblocco è quello di minix v3
			memcpy(sp, buf, sizeof(minix3_super_block));
			magic = sp->s_magic;
			flog(LOG_DEBUG,"versione 3");
			version = V3;
			native = 1; 
		} else {
			mem_free(buf);
			return(EINVAL);
		}
	}

	mem_free(buf);

	/* If the super block has the wrong byte order, swap the fields; the magic
	* number doesn't need conversion. */
	sp->s_ninodes =       conv4(native, sp->s_ninodes);
	sp->s_nzones =        conv2(native, (int) sp->s_nzones);
	sp->s_imap_blocks =   conv2(native, (int) sp->s_imap_blocks);
	sp->s_zmap_blocks =   conv2(native, (int) sp->s_zmap_blocks);
	sp->s_firstdatazone = conv2(native, (int) sp->s_firstdatazone);
	sp->s_log_zone_size = conv2(native, (int) sp->s_log_zone_size);
	sp->s_max_size =      conv4(native, sp->s_max_size);
	sp->s_zones =         conv4(native, sp->s_zones);
	sp->s_disk_version =  version; // serve quando vado a leggere le cartelle

	flog(LOG_DEBUG, "sp->s_ninodes:%d",sp->s_ninodes);
	flog(LOG_DEBUG, "sp->s_nzones:%d",sp->s_nzones);
	flog(LOG_DEBUG, "sp->s_imap_blocks:%d",sp->s_imap_blocks);
	flog(LOG_DEBUG, "sp->s_zmap_blocks:%d",sp->s_zmap_blocks);
	flog(LOG_DEBUG, "sp->s_firstdatazone:%d",sp->s_firstdatazone);
	flog(LOG_DEBUG, "sp->s_log_zone_size:%d",sp->s_log_zone_size);
	flog(LOG_DEBUG, "sp->s_max_size:%d",sp->s_max_size);
	flog(LOG_DEBUG, "sp->s_zones:%d",sp->s_zones);
	flog(LOG_DEBUG, "sp->s_disk_version:%d",sp->s_disk_version);


	/* In V1, the device size was kept in a short, s_nzones, which limited
	* devices to 32K zones.  For V2, it was decided to keep the size as a
	* long.  However, just changing s_nzones to a long would not work, since
	* then the position of s_magic in the super block would not be the same
	* in V1 and V2 file systems, and there would be no way to tell whether
	* a newly mounted file system was V1 or V2.  The solution was to introduce
	* a new variable, s_zones, and copy the size there.
	*
	* Calculate some other numbers that depend on the version here too, to
	* hide some of the differences.
	*/
	if (version == V1 || version == V1_REV) {
		sp->s_block_size = BLOCK_SIZE;
		sp->s_zones = sp->s_nzones; /* only V1 needs this copy */
		sp->s_inodes_per_block = V1_INODES_PER_BLOCK;
		sp->s_ndzones = V1_NR_DZONES;
		sp->s_nindirs = V1_INDIRECTS;
	} else {
		if (version == V2 || version == V2_REV)
			sp->s_block_size = BLOCK_SIZE;
		if (sp->s_block_size < BLOCK_SIZE)
			return EINVAL;
		sp->s_inodes_per_block = V2_INODES_PER_BLOCK(sp->s_block_size);
		sp->s_ndzones = V2_NR_DZONES;
		sp->s_nindirs = V2_INDIRECTS(sp->s_block_size);
	}

	if (sp->s_block_size < BLOCK_SIZE) {
		return EINVAL;
	}
	
	if (sp->s_block_size > BLOCK_SIZE) {
		flog(LOG_DEBUG,"Filesystem block size is %d kB; maximum filesystem\n"
		"block size is %d kB. This limit can be increased by recompiling.\n",
		sp->s_block_size/1024, BLOCK_SIZE/1024);
		return EINVAL;
	}
	if ((sp->s_block_size % 512) != 0) {
		return EINVAL;
	}
	if (SUPER_SIZE > sp->s_block_size) {
		return EINVAL;
	}

	if ((sp->s_block_size % V2_INODE_SIZE) != 0 || (sp->s_block_size % V1_INODE_SIZE) != 0) {
		return EINVAL;
	}

	sp->s_isearch = 0;    // inode searches initially start at 0 
	sp->s_zsearch = 0;    // zone searches initially start at 0 
	sp->s_version = version;
	sp->s_native  = native;

	// Make a few basic checks to see if super block looks reasonable. 
	if (sp->s_imap_blocks < 1 || sp->s_zmap_blocks < 1
	   || sp->s_ninodes < 1 || sp->s_zones < 1
	   || (unsigned) sp->s_log_zone_size > 4) {
		flog(LOG_ERR,"not enough imap or zone map blocks,");
		flog(LOG_ERR,"or not enough inodes, or not enough zones, or zone size too large");
		return(EINVAL);
	}
	sp->s_dev = dev;    // restore device number 
	return(OK);
}
