/* This file is the counterpart of "read.c".  It contains the code for writing
 * insofar as this is not contained in read_write().
 *
 * The entry points into this file are
 *   do_write:     call read_write to perform the WRITE system call
 *   clear_zone:   erase a zone in the middle of a file
 *   new_block:    acquire a new block
 */

#include "libce.h"
#include "fs.h"
//#include <string.h>
//#include "buf.h"
#include "file.h"
#include "fproc.h"
#include "inode.h"
#include "super.h"
#include "glo.h"

PRIVATE int write_map(inode *rip, off_t position, zone_t new_zone);
PRIVATE void wr_indir(blocco_fis *bp, int index, zone_t zone);

/*===========================================================================*
 *				do_write				     *
 *===========================================================================*/
PUBLIC
int do_write(int fd, void *buf, uint32_t nbytes)
{
// Perform the write(fd, buffer, nbytes) system call.
/* Soluzione provvisoria:
	uso un semaforo per la mutua esclusione. 
	Può rallentare di molto l'esecuzione */	
	int res;
	// la memoria allocata nello heap è residente 
	// ed inoltre è nello spazio condiviso
	void *local_buf = mem_alloc(nbytes);
	memcpy(local_buf,buf,nbytes);
	res = read_write(WRITING,fd,(char*)buf,nbytes);
	mem_free(local_buf);
	return res;
}

/*===========================================================================*
 *				write_map				     *
 *===========================================================================*/
PRIVATE int write_map(inode *rip, off_t position, zone_t new_zone)
//register struct inode *rip;	/* pointer to inode to be changed */
//off_t position;			/* file address to be mapped */
//zone_t new_zone;		/* zone # to be inserted */
{
/* Write a new zone into an inode. */
	int scale, ind_ex, new_ind, new_dbl, zones, nr_indirects, single, zindex, ex;
	zone_t z, z1;
	block_t b;
	int excess, zone;
	//struct buf *bp;
	blocco_fis *bp = NIL_BLK;

	rip->i_dirt = DIRTY;		// inode will be changed 
	//bp = NIL_BUF;
	scale = rip->i_sp->s_log_zone_size;		/* for zone-block conversion */
	/* relative zone # to insert */
	zone = (position/rip->i_sp->s_block_size) >> scale;
	zones = rip->i_ndzones;	/* # direct zones in the inode */
	nr_indirects = rip->i_nindirs;/* # indirect zones per indirect block */

	/* Is 'position' to be found in the inode itself? */
	if (zone < zones) {
		zindex = (int) zone;	/* we need an integer here */
		rip->i_zone[zindex] = new_zone;
		flog(LOG_DEBUG,"write_map zindex:%d new_zone:%d",zindex,new_zone);
		return(OK);
	}

	// It is not in the inode, so it must be single or double indirect. 
	excess = zone - zones;	// first Vx_NR_DZONES don't count 
	new_ind = FALSE;
	new_dbl = FALSE;

	if (excess < nr_indirects) {
		// 'position' can be located via the single indirect block. 
		z1 = rip->i_zone[zones];	// single indirect zone 
		single = TRUE;
	} else {
	  	// 'position' can be located via the double indirect block. 
	  	if ( (z = rip->i_zone[zones+1]) == NO_ZONE) {
	  		// Create the double indirect block. 
	  		if ( (z = alloc_zone(rip->i_dev, rip->i_zone[0])) == NO_ZONE)
	  			return(err_code);
	  		rip->i_zone[zones+1] = z;
	  		new_dbl = TRUE;	// set flag for later 
	  	}

		// Either way, 'z' is zone number for double indirect block. 
		excess -= nr_indirects;	// single indirect doesn't count 
		ind_ex = (int) (excess / nr_indirects);
		excess = excess % nr_indirects;
		if (ind_ex >= nr_indirects) return(EFBIG);
		b = (block_t) z << scale;
		//bp = get_block(rip->i_dev, b, (new_dbl ? NO_READ : NORMAL));
		bp = get_block(rip->i_dev,b,NORMAL);
		if (new_dbl) zero_block(bp);
		z1 = rd_indir(bp, ind_ex);
		single = FALSE;
	}

	// z1 is now single indirect zone; 'excess' is index. 
	if (z1 == NO_ZONE) {
		// Create indirect block and store zone # in inode or dbl indir blk. 
		z1 = alloc_zone(rip->i_dev, rip->i_zone[0]);
		if (single)
			rip->i_zone[zones] = z1;	// update inode 
		else
			wr_indir(bp, ind_ex, z1);	// update dbl indir 

		new_ind = TRUE;
		//if (bp != NIL_BUF) bp->b_dirt = DIRTY;	// if double ind, it is dirty
		if (z1 == NO_ZONE) {
			//put_block(bp, INDIRECT_BLOCK);	// release dbl indirect blk 
			put_block(rip->i_dev,b, bp);
			return(err_code);	// couldn't create single ind 
		}
	}
	//put_block(bp, INDIRECT_BLOCK);	// release double indirect blk 
	put_block(rip->i_dev, b, bp);

	// z1 is indirect block's zone number. 
	b = (block_t) z1 << scale;
	bp = get_block(rip->i_dev, b, (new_ind ? NO_READ : NORMAL) );
	if (new_ind) zero_block(bp);
	ex = (int) excess;			// we need an int here 
	wr_indir(bp, ex, new_zone);
	//bp->b_dirt = DIRTY;
	//put_block(bp, INDIRECT_BLOCK); 
	//c_dmawritehd_n((uint16_t*)bp,b,1,errore);
	put_block(rip->i_dev, b, bp);
	return(OK);
}

/*===========================================================================*
 *				wr_indir				     *
 *===========================================================================*/
PRIVATE void wr_indir(blocco_fis *bp, int index, zone_t zone)
//struct buf *bp;			/* pointer to indirect block */
//int index;			/* index into *bp */
//zone_t zone;			/* zone to write */
{
/* Given a pointer to an indirect block, write one entry. */

	super_block *sp = get_super(CE_DEV);
	// sp = get_super(bp->b_dev);	

  // write a zone into an indirect block 
	if (sp->s_version == V1)
		bp->b_v1_ind[index] = (zone1_t) conv2(sp->s_native, (uint16_t)  zone);
	else
		bp->b_v2_ind[index] = (zone_t)  conv4(sp->s_native, (uint32_t) zone);
}

/*===========================================================================*
 *				clear_zone				     *
 *===========================================================================*/
PUBLIC void clear_zone(inode *rip, off_t pos, int flag)
//register struct inode *rip;	/* inode to clear */
//off_t pos;			/* points to block to clear */
//int flag;			/* 0 if called by read_write, 1 by new_block */
{
/* Zero a zone, possibly starting in the middle.  The parameter 'pos' gives
 * a byte in the first block to be zeroed.  Clearzone() is called from 
 * read_write and new_block().
 */

  //register struct buf *bp;
	blocco_fis *bp;
	block_t b, blo, bhi;
	off_t next;
	int scale;
	zone_t zone_size;

  /* If the block size and zone size are the same, clear_zone() not needed. */
  	scale = rip->i_sp->s_log_zone_size;
  	if (scale == 0) return;

	zone_size = (zone_t) rip->i_sp->s_block_size << scale;
	if (flag == 1) 
		pos = (pos/zone_size) * zone_size;
	next = pos + rip->i_sp->s_block_size - 1;

  	/* If 'pos' is in the last block of a zone, do not clear the zone. */
	if (next/zone_size != pos/zone_size) return;
	if ( (blo = read_map(rip, next)) == NO_BLOCK) return;
	bhi = ( ((blo>>scale)+1) << scale ) - 1;

  	/* Clear all the blocks between 'blo' and 'bhi'. */
  	for (b = blo; b <= bhi; b++) {
		bp = get_block(rip->i_dev, b, NO_READ);
		zero_block(bp);
		//put_block(bp, FULL_DATA_BLOCK);
		//c_dmawritehd_n((uint16_t*)bp,b,1,errore);
		put_block(rip->i_dev, b, bp);
  	}
}

/*===========================================================================*
 *				new_block				     *
 *===========================================================================*/
PUBLIC blocco_fis *new_block(inode *rip, off_t position)
//register struct inode *rip;	/* pointer to inode */
//off_t position;			/* file pointer */
{
/* Acquire a new block and return a pointer to it.  Doing so may require
 * allocating a complete zone, and then returning the initial block.
 * On the other hand, the current zone may still have some unused blocks.
 */

	blocco_fis *bp;
	block_t b, base_block;
	zone_t z;
	zone_t zone_size;
	int scale, r;
	super_block *sp;

 	// Is another block available in the current zone? 
  	if ( (b = read_map(rip, position)) == NO_BLOCK) {
		// Choose first zone if possible. 
		/* Lose if the file is nonempty but the first zone number is NO_ZONE
		* corresponding to a zone full of zeros.  It would be better to
		* search near the last real zone.
		*/
		if (rip->i_zone[0] == NO_ZONE) {
			//flog(LOG_DEBUG,"new_block rip->i_zone[0]==NO_ZONE rip->i_dev:%d s_firstdatazone:%d",rip->i_dev,rip->i_sp->s_firstdatazone);
			sp = rip->i_sp;
			z = sp->s_firstdatazone;
		} else {
			z = rip->i_zone[0];	/* hunt near first zone */
		}
		if ( (z = alloc_zone(rip->i_dev, z)) == NO_ZONE) {
			//flog(LOG_DEBUG,"new_block alloc_zone fallita, z:%d",z);			
			return NIL_BLK;//return(NIL_BUF);
		}
		if ( (r = write_map(rip, position, z)) != OK) {
			free_zone(rip->i_dev, z);
			err_code = r;
			//return(NIL_BUF);
			return NIL_BLK;
		}
		//flog(LOG_DEBUG,"new_block rip->i_zone[0]:%d",rip->i_zone[0]);
		/* If we are not writing at EOF, clear the zone, just to be safe. */
		if ( position != rip->i_size) clear_zone(rip, position, 1);
		scale = rip->i_sp->s_log_zone_size;
		base_block = (block_t) z << scale;
		zone_size = (zone_t) rip->i_sp->s_block_size << scale;
		b = base_block + (block_t)((position % zone_size)/rip->i_sp->s_block_size);
		//flog(LOG_DEBUG,"new_block b:%d",b);
  	}

	bp = get_block(rip->i_dev, b, NO_READ);
	zero_block(bp);
	// non esiste la cache
	put_block(rip->i_dev, b, bp);
	return(bp);
}

/*===========================================================================*
 *				zero_block				     *
 *===========================================================================*/
 // azzera un blocco fisico
PUBLIC void zero_block(blocco_fis *bp)
//register struct buf *bp;	// pointer to buffer to zero 
{
	// Zero a block. 
	memset(bp->b_data, 0, BLOCK_SIZE);
	//bp->b_dirt = DIRTY;
}
